package com.engidea;

import com.engidea.comm.CommPort;
import com.engidea.comm.CommPortIdentifier;
import com.engidea.comm.SerialPort;
import com.engidea.comm.SerialPortEvent;
import com.engidea.comm.SerialPortEventListener;
import com.engidea.linjcom.LinjcomIdentifier;
import com.engidea.win32jcom.WinjcomIdentifier;

import java.io.*;
import java.util.Iterator;
import java.util.List;

/*
  WinJcom is a native interface to serial ports in java.
  Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License
 */

/**
 * Simple class that can list system ports and allow IO
 */
public class PortTester 
  {
  public static void main(String[] args) 
    {
    PortTester portTester = new PortTester();
    portTester.StartTest();
    }

  private CommPortIdentifier portIdentifier;
  private SerialPort serport;

  private void report ( String msg )
    {
    System.out.println(msg);
    }
  
  private CommPortIdentifier newComPortIdentifier (int debugMask)
    {
    String os_name = System.getProperty("os.name");
    
    if ( os_name == null || os_name.indexOf("Linux") >= 0 )
      return new LinjcomIdentifier(debugMask);
    else
      return new WinjcomIdentifier(debugMask);
    }
  
  
  private void StartTest() 
    {
    portIdentifier = newComPortIdentifier(0x1);
    listComPorts();
    selectComport();
    new Thread(new PortReader()).start();
    typeSendBytes();
    }

  private void typeSendBytes()
    {  
    try
      {
      BufferedReader in = new BufferedReader(new InputStreamReader(System.in));
      String aStr = "";
      while (aStr != null) 
        {
        System.out.print("type something to send> ");
        aStr = in.readLine();
        aStr += "\r\n";
        // WARNING: be careful, you should select the encoding !
        // This will timeout if you have FLOW CONTROL and the line is stuck !
        byte []buf = aStr.getBytes();
        serport.write(buf,0,buf.length );
        }
      }
    catch ( Exception exc )
      {
      exc.printStackTrace();
      }
    }
    
  private void listComPorts ()
    {
    List<CommPort> portlist = portIdentifier.getCommPortList();
    for (Iterator<CommPort> iter=portlist.iterator(); iter.hasNext(); )
      {
      CommPort identifier = iter.next();
      System.out.println("Available port name="+identifier.getName());
      }
    }

  private SerialPort openPort ( String portName )
    {
    try 
      {
      CommPort aPort = portIdentifier.getCommPort(portName);
      aPort.open();
      return (SerialPort)aPort;
      }
    catch ( Exception exc )
      {
      report("exc="+exc);
      exc.printStackTrace();
      }
      
    return null;
    }
    
  
  private void sleep (int seconds )
    {
    try
      {
      Thread.sleep(seconds*1000);
      }
    catch ( Exception exc )
      {
      exc.printStackTrace();
      }
    }
  
  
  private void testDTR ( SerialPort serport) throws IOException
    {
    serport.setDTR(true);
    sleep(2);
    serport.setDTR(false);
    sleep(2);
    serport.setDTR(true);
    sleep(2);
    serport.setDTR(false);
    }
  
  private void testRTS ( SerialPort serport) throws IOException
    {
    // to test RTS you have to disable flow control
    serport.setFlowControlMode(SerialPort.FLOWCONTROL_NONE);

    serport.setRTS(false);
    sleep(2);
    serport.setRTS(true);
    sleep(2);
    serport.setRTS(false);
    sleep(2);
    serport.setRTS(true);
    } 

  private void selectComport ()
    {
    try 
      {
      BufferedReader in = new BufferedReader(new InputStreamReader(System.in));

      System.out.print("type port name> ");
      String portname = in.readLine();
      serport = openPort(portname);
      serport.setSerialPortParams(9600,8, SerialPort.STOPBITS_2, SerialPort.PARITY_NONE);
      serport.enableReceiveTimeout(10000);
      serport.setEventListener(new EventListener());

      serport.notifyOnDSR(true);
      serport.notifyOnCarrierDetect(true);
      
      report("have baudrate="+serport.getBaudRate());
      report("isDTR()="+serport.isDTR());
      
//      testDTR(serport);
//      testRTS(serport);
      } 
    catch (IOException exc) 
      {
      report("Exc="+exc);
      exc.printStackTrace();
      }
    }


private final class EventListener implements SerialPortEventListener
  {
  public void serialEvent(SerialPortEvent ev)
    {
    System.out.println("Got event="+ev);
    }
  }


private final class PortReader implements Runnable
  {
  
  private void readSomething () throws Exception
    {
    // This will timeout if nothing is received in the specified time.
    byte []buff = new byte[5];
    
    int letti = serport.read(buff,0,buff.length);
    
    if ( letti <= 0 )
      {
      report ("read: timeout");
      return;
      }
    
    String toprint = new String(buff, 0, letti);
    
    System.out.print(toprint);
    }
  
  public void run()
    {
    report ("PortReader: run()");
    
    try
      {
      for (;;) readSomething();
      }
    catch ( Exception exc )
      {
      exc.printStackTrace();
      }
    
    report ("PortReader: end()");
    }
  }


}
