package com.engidea.comm;


import java.io.*;

/*
  WinJcom is a native interface to serial ports in java.
  Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
*/

/**
 * The main difference are the method read and write.
 * They are needed since it may happen that a port returns less bytes than wanted or writes less
 * bytes that you are expected to write. read and write tells you this.
 * If you want the standard interface just use PortInputStream and PortOutputStream
 * NOTE: Basically all methods return either IOexception or UnsupportedEsception, this
 * is really by design since IO is a messy business and you MUST NOT assume things go smooth and nice.
 * methods may fail for a multitude of reasons, locking, ownership, timeout, BUGS, you HAVE
 * to take care of exception either in the form or reopen or in user notification.
 */
public abstract class CommPort 
  {
  protected final String portName;
  
  private final int debugMask;
  
  /**
   * Constructor, needed since portName is final.
   * @param portName
   */
  public CommPort ( String portName, int debugMask )
    {
    this.portName  = portName;
    this.debugMask = debugMask;
    }
    
  /**
   * Accessor for portName
   * @return the name of the port as set in costructor
   */
  public String getName()
    {
    return portName;
    }

  protected void report ( String msg )
    {
    System.out.println(msg);
    }
  
  protected boolean mustLog ( int mask )
    {
    return (debugMask & mask) != 0;
    }
  
  
  public abstract void open() throws IOException;
  
  public abstract void close();

  /**
   * Port instances are allowed to read less chars than requested at their wish.
   * Expect this to throw exception on timeout or alternatively to return zero bytes.
   * @param buff
   * @param offset
   * @param len 
   * @return the number of bytes read, only -1 means EOF, zero is a valid return
   * @throws IOException
   */
  public abstract int read ( byte[] buff, int offset, int len) throws IOException;  

  /**
   * Port instances are allowed to write less chars than requested.
   * Expect this to throw exception on timeout or alternatively to return zero bytes.
   * @param buff
   * @param offset
   * @param len
   * @return the number of bytes written, only -1 means EOF, Zero is a valid return
   * @throws IOException
   */
  public abstract int write ( byte[]buff, int offset, int len ) throws IOException;

  /**
   * Return the number of bytes available on read.
   * @return the number of bytes available to read
   * @throws IOException
   */
  public abstract int available() throws IOException;
  
  /**
   * This THROWS AWAY any char that is in the input or output queue !!!!
   * @throws IOException
   */
  public abstract void flush() throws IOException;






  public abstract void enableReceiveFraming(int f) throws IOException;

  public abstract void disableReceiveFraming() throws IOException;

  public abstract boolean isReceiveFramingEnabled() throws IOException;

  public abstract int getReceiveFramingByte() throws IOException;

  public abstract void disableReceiveTimeout() throws IOException;

  public abstract void enableReceiveTimeout(int timeoutMilli) throws IOException;

  public abstract boolean isReceiveTimeoutEnabled() throws IOException;

  public abstract int getReceiveTimeout() throws IOException;

  public void enableReceiveThreshold(int thresh)throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public void disableReceiveThreshold() throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public int getReceiveThreshold() throws IOException
    {    
    throw new UnsupportedException("Not implemented");
    }

  public boolean isReceiveThresholdEnabled() throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }
  
  public void setInputBufferSize(int size)throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public int getInputBufferSize()throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public void setOutputBufferSize(int size)throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public int getOutputBufferSize()throws IOException
    {
    throw new UnsupportedException("Not implemented");
    }

  public String toString()
    {
    return portName;
    }
  }
