package com.engidea.linjcom;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.engidea.comm.CommPort;
import com.engidea.comm.CommPortIdentifier;


/*
LinJcom is a native interface to serial ports in java.
Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License
*/ 


public final class LinjcomIdentifier extends CommPortIdentifier
  {
  private static final String classname="LinjcomIdentifier.";

  // this debug mask is used even in the native part of the driver
  // if you change the values you MUST recreate the library
  public static final int DEBUG_INIT       = 0x000000001;
  public static final int DEBUG_TEST_PORT  = 0x000000002;
  public static final int DEBUG_ERRORS     = 0x000000004;    
  public static final int DEBUG_OPEN_PORT  = 0x000000008;   
  public static final int DEBUG_WRITE      = 0x000000010;   
  public static final int DEBUG_READ       = 0x000000020;   
  public static final int DEBUG_COMMEVENT  = 0x000000040;   
  public static final int DEBUG_POPARAMS   = 0x000000080;   
  public static final int DEBUG_CLOSE_PORT = 0x000000100;

  
  /**
   * You start with this one
   * You can enable debugging mask if you wish so.
   * @param debugMask a debug mask taken from the above constants. Zero means no debug
   */
  public LinjcomIdentifier (int debugMask )
    {
    this(debugMask, guessNativeLibrary());
    }

  private static String guessNativeLibrary ()
    {
    String os_arch = System.getProperty("os.arch");
    
    if ( os_arch == null || os_arch.indexOf("amd64") >= 0 )
    	return "jcom64";
    else
    	return "jcom";
    }
  
  /**
   * Use this one If the system cannot reliably detect if it is on vista or not
   * You can enable debugging mask if you wish so.
   * @param debugMask a debug mask taken from the above constants. Zero means no debug
   * @param native_library the native library to be used to support java
   */
  public LinjcomIdentifier (int debugMask, String native_library)
    {
    this.debugMask = debugMask;

    if (mustLog(DEBUG_INIT) )
      {
      report (classname+"initialize native lib=" + native_library);
      report ("os.name="+System.getProperty("os.name"));
      report ("os.arch="+System.getProperty("os.arch"));
      }
     
    try
      {
      Runtime runtime = Runtime.getRuntime();
      
      // load native library
      runtime.loadLibrary(native_library);
      }
    catch ( Throwable exc )
      {
      exc.printStackTrace();
      
      System.err.println("java.library.path:");

      String property = System.getProperty("java.library.path");
      String []items = property.split(":");
      
      for (int index=0; index<items.length; index++)
          System.err.println("  "+items[index]);

      }
    
    // Initialize the native part of the library
    nativeInitialize(debugMask);
    }
    

  private native void nativeInitialize(int debugMaks);

  /**
   * Note that his one does NOT leave the port "open"
   * If test is OK then it returns true, false + exception otherwise
   * @param dev
   * @throws IOException
   */
  private native boolean nativeIsPortPresent(String dev) throws IOException;

  private boolean isPortPresent( String portName )
    {
    try
      {
      // this will throw exception if this port is not available.
      return nativeIsPortPresent(portName);
      }
    catch ( Exception exc )
      {
      if (mustLog(DEBUG_TEST_PORT)) report (classname+"isPortPresent() port="+portName+" exc="+exc);
      return false;
      }
    }


   @Override
   public CommPort getCommPort(String portName)
    {
    return new LinjcomPort(portName,debugMask);
    }

   @Override
   public List<CommPort> getCommPortList()
    {
    return getCommPortList ( 20 );
    }
  
  @Override
  public List<CommPort> getCommPortList(int maxPortIndex)
    {
    ArrayList<CommPort> risul = new ArrayList<CommPort>();
    
    for (int index=0; index<maxPortIndex; index++)
      {
      String portName = "/dev/ttyACM"+index;
      
      if ( ! isPortPresent(portName)) continue;
      
      risul.add(getCommPort(portName));
      }
    
    for (int index=0; index<maxPortIndex; index++)
      {
      String portName = "/dev/ttyUSB"+index;
      
      if ( ! isPortPresent(portName)) continue;
      
      risul.add(getCommPort(portName));
      }

    
    return risul;
    }
  }
