package com.engidea.linjcom;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import com.engidea.comm.CommPort;
import com.engidea.comm.CommPortIdentifier;
import com.engidea.win32jcom.WinjcomPort;

/*
LinJcom is a native interface to serial ports in java.
Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License
*/


public final class Linjcomidentifier extends CommPortIdentifier
  {
  private static final String classname="Linjcomidentifier.";

  public static final String  jcomLinuxLibrary="jcom";
  
  // this debug mask is used even in the native part of the driver
  // if you change the values you MUST recreate the dll.
  public static final int DEBUG_INIT       = 0x000000001;
  public static final int DEBUG_TEST_PORT  = 0x000000002;
  public static final int DEBUG_ERRORS     = 0x000000004;    
  public static final int DEBUG_OPEN_PORT  = 0x000000008;   
  public static final int DEBUG_WRITE      = 0x000000010;   
  public static final int DEBUG_READ       = 0x000000020;   
  public static final int DEBUG_COMMEVENT  = 0x000000040;   
  public static final int DEBUG_POPARAMS   = 0x000000080;   
  public static final int DEBUG_CLOSE_PORT = 0x000000100;

  
  /**
   * You start with this one
   * You can enable debugging mask if you wish so.
   * @param debugMask a debug mask taken from the above constants. Zero means no debug
   */
  public Linjcomidentifier (int debugMask )
    {
    this(debugMask, guessNativeLibrary());
    }

  private static String guessNativeLibrary ()
    {
    String os_name = System.getProperty("os.name");
    
    return jcomLinuxLibrary;
    }
  
  /**
   * Use this one If the system cannot reliably detect if it is on vista or not
   * You can enable debugging mask if you wish so.
   * @param debugMask a debug mask taken from the above constants. Zero means no debug
   * @param native_library the native library to be used to support java
   */
  public Linjcomidentifier (int debugMask, String native_library)
    {
    this.debugMask = debugMask;

    if (mustLog(DEBUG_INIT) ) report (classname+"initialize native lib=" + native_library);
     
    try
      {
      Runtime runtime = Runtime.getRuntime();
      
      // load native library
      runtime.loadLibrary(native_library);
      }
    catch ( Throwable exc )
      {
      exc.printStackTrace();
      
      System.err.println("java.library.path:");

      String property = System.getProperty("java.library.path");
      String []items = property.split(":");
      
      for (int index=0; index<items.length; index++)
          System.err.println("  "+items[index]);

      }
    
    // Init the native part of the library
    nativeInitialize(debugMask);
    }
    
  private boolean mustLog ( int mask )
    {
    return (debugMask & mask) != 0;
    }
    
  private void report ( String msg )
    {
    System.out.println(msg);
    }

  private native void nativeInitialize(int debugMaks);

  /**
   * Note that his one does NOT leave the port "open"
   * If test is OK then it returns true, false + exception othervise
   * @param dev
   * @throws IOException
   */
  private native boolean nativeIsPortPresent(String dev) throws IOException;

  private boolean isPortPresent( String portName )
    {
    try
      {
      // this will throw exception if this port is not available.
      return nativeIsPortPresent(portName);
      }
    catch ( Exception exc )
      {
      if (mustLog(DEBUG_TEST_PORT)) report (classname+"isPortPresent() port="+portName+" exc="+exc);
      return false;
      }
    }


  /**
   * See the API description for documentation.
   * @param portName
   * @return usable com port attached to the given name
   */
  public CommPort getCommPort(String portName)
    {
    return new WinjcomPort(portName,debugMask);
    }

  /**
   * The old api, adjusted to scan up to 20ports
   * @return
   */
  public List<CommPort> getCommPortList()
    {
    return getCommPortList ( 20 );
    }

  /**
   * See at API description for documentation.
   * Windows com port are named COM1 .... COM9 for the first 9 ports then Billgates
   * decided that it was too easy and the following ones will be named \\.\com10
   * Do not ask me why, but this is how it is
   * @return a List of CommPort present on the system.
   */
  public List<CommPort> getCommPortList(int maxPortIndex)
    {
    ArrayList<CommPort> risul = new ArrayList<CommPort>();
    
    for (int index=1; index<10; index++)
      {
      String portName = "COM"+index;
      
      if ( !isPortPresent(portName)) continue;
      
      risul.add(getCommPort(portName));
      }
    
    for (int index=10; index<maxPortIndex; index++)  
      {
      // Contributed by "John Muczynski"
      String portName = "\\\\.\\com"+index;

      if ( !isPortPresent(portName)) continue;

      risul.add(getCommPort(portName));
      }    
    
    return risul;
    }
  }
