/**
  WinJcom is a native interface to serial ports in java.
  Copyright 2007 by Damiano Bolla, Jcomm@engidea.com

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License
*/


#include <termios.h>
#include <unistd.h>
#include <string.h>

#include <sys/types.h>  // open
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <errno.h>

#include <stdio.h>   // select
#include <stdlib.h>
#include <sys/time.h>

// Include JNI prototyped
#include "jcomport.h"


#define IO_EXCEPTION "java/io/IOException"
#define ILLEGAL_EXCEPTION "java/lang/IllegalArgumentException"


#define DEBUG_INIT       com_engidea_linjcom_LinjcomIdentifier_DEBUG_INIT
#define DEBUG_TEST_PORT  com_engidea_linjcom_LinjcomIdentifier_DEBUG_TEST_PORT
#define DEBUG_ERRORS     com_engidea_linjcom_LinjcomIdentifier_DEBUG_ERRORS
#define DEBUG_OPEN_PORT  com_engidea_linjcom_LinjcomIdentifier_DEBUG_OPEN_PORT
#define DEBUG_WRITE      com_engidea_linjcom_LinjcomIdentifier_DEBUG_WRITE
#define DEBUG_READ       com_engidea_linjcom_LinjcomIdentifier_DEBUG_READ
#define DEBUG_COMMEVENT  com_engidea_linjcom_LinjcomIdentifier_DEBUG_COMMEVENT
#define DEBUG_POPARAMS   com_engidea_linjcom_LinjcomIdentifier_DEBUG_POPARAMS
#define DEBUG_CLOSE_PORT com_engidea_linjcom_LinjcomIdentifier_DEBUG_CLOSE_PORT

#define ERROR_NULL	 0
#define TRUE		1
#define FALSE		0
#define NAME_MAX	128
#define DS_COM_TIMEOUT_MAX  255     // VTIME is in tenth of seconds and is up to 255

typedef unsigned int uint32;

typedef int int32;

static jint debugMask;

struct ComStatus
   {
   void *myselfPtr;             // pointer to myself to check for validity status
   int  portFd;                 // Linux port file descriptor
   struct termios termios;      // store current port configuration to avoid getting it often
   uint32 timeoutDs;            // IO timeout to be set, both RX and TX in tenth of seconds
   struct timeval writeTimeval; // prepared timeval ready to be used for write select
   uint32 notifyEnableBits;     // Bits TIOCM_RNG, TIOCM_DSR, TIOCM_CD, and TIOCM_CTS.
   };



/**
 * @return true if I must log this line
 */
static int mustLog ( int mask )
   {
   return debugMask & mask;
   }


/**
 * This is the first one being called after library is loaded
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomIdentifier_nativeInitialize (JNIEnv *env, jobject jobj, jint jmask)
   {
   debugMask = jmask;

   if ( mustLog(DEBUG_INIT) ) printf ("nativeInitialize: debugMaks=0x%X\n",debugMask);
   }


static void throw_exception( JNIEnv *env, char *exc, char *func, char *msg, int32 lastError )
   {
   char buf[1000];
   char *errorStr;
   
   jthrowable previous_exc = (*env)->ExceptionOccurred(env);
   
   if ( previous_exc )
      {
      // If there is an exception logged, leave this active, it is probably more relevant
	  return;
      }
   
   jclass javaclass = (*env)->FindClass( env, exc );

   if( javaclass == NULL ) 
      {
      // this also should never happens....
      printf ("throw_exception: ERROR: cannot get exception class [%s]\n",exc);
      return;
      }
   
   if ( lastError )
      {
	  if ( (errorStr = strerror(lastError)) )
		  snprintf( buf,sizeof(buf), "fun(%s) msg(%s) err(%d:%s)",func,msg,lastError,errorStr);
	  else
		  snprintf( buf,sizeof(buf), "fun(%s) msg(%s) err(%d)",func,msg,lastError);
      }
   else
      {
      snprintf( buf,sizeof(buf), "fun(%s) msg(%s)",func,msg);
      }

   (*env)->ThrowNew( env, javaclass, buf );
   (*env)->DeleteLocalRef( env, javaclass );
   }


#pragma GCC diagnostic ignored "-Wpointer-to-int-cast"
#pragma GCC diagnostic ignored "-Wint-to-pointer-cast"


static struct ComStatus * validateComPtr (JNIEnv *env, jlong comPtr )
   {
   if ( comPtr == 0 )
      {
   	  throw_exception( env, ILLEGAL_EXCEPTION, "validateComPtr", "comPtr is NULL", ERROR_NULL );
   	  return NULL;
      }

   struct ComStatus *comP = (struct ComStatus *)comPtr;

   if ( comP->myselfPtr != comP )
      {
   	  throw_exception( env, ILLEGAL_EXCEPTION, "validateComPtr", "comPtr validate FAIL", ERROR_NULL );
   	  return NULL;
      }

   return comP;
   }


/**
 * From http://www.club.cc.cmu.edu/~cmccabe/blog_jni_flaws.html
 */
static int jstr_to_cstr(JNIEnv *env, jstring jstr, char *cstr, int cstr_len)
	{
	if ( jstr == NULL )
       {
       throw_exception( env, IO_EXCEPTION, "jstr_to_cstr", "name_len == NULL", ERROR_NULL );
       return -ESTALE;
       }

    int clen = (*env)->GetStringUTFLength(env, jstr);

    if (clen > cstr_len)
       {
       throw_exception( env, IO_EXCEPTION, "jstr_to_cstr", "name_len too long", ERROR_NULL );
       return -ENAMETOOLONG;
       }

    int jlen = (*env)->GetStringLength(env, jstr);

    (*env)->GetStringUTFRegion(env, jstr, 0, jlen, cstr);

    if ((*env)->ExceptionCheck(env))
       {
       throw_exception( env, IO_EXCEPTION, "jstr_to_cstr", "IO error", ERROR_NULL );
       return -ESTALE;
       }

    return 0;
	}

/**
 * return a valid Linux file descriptor or < 1 and an error if it fails
 */
static int open_com_path ( JNIEnv *env, jstring port_name )
   {
   char portName[NAME_MAX+2];

   if ( jstr_to_cstr (env, port_name, portName, NAME_MAX ) ) return -1;

   if ( mustLog(DEBUG_OPEN_PORT)) printf("open_com_path: name(%s) \n",portName);

   // avoid getting stuck on opening the port
   return open (portName, O_RDWR | O_NOCTTY | O_SYNC | O_NONBLOCK );
   }

/**
 * Test if a given native port exists or not
 */
JNIEXPORT jboolean JNICALL Java_com_engidea_linjcom_LinjcomIdentifier_nativeIsPortPresent (JNIEnv *env, jobject jobj, jstring port_name )
   {
   int comFd = open_com_path (env, port_name );

   if ( comFd < 0 )
      {
      throw_exception(env,IO_EXCEPTION,"isPortPresent","open FAIL",errno);
      return FALSE;
      }

   // From now on need to close before return

   if ( ! isatty(comFd) )
      {
      throw_exception(env,IO_EXCEPTION,"isPortPresent","NOT a tty", ERROR_NULL);
      close (comFd);
      return FALSE;
      }

   close (comFd);
   return TRUE;
   }



static int setNonblock (struct ComStatus *comP, int enable )
   {
   int flags = fcntl(comP->portFd, F_GETFL, 0);

   if ( enable )
	  return fcntl(comP->portFd, F_SETFL, flags | O_NONBLOCK );
   else
	   return fcntl(comP->portFd, F_SETFL, flags & ~O_NONBLOCK );
   }



/**
 * Open a port to be used.
 * @return a handle to an opaque structure or NULL(0) if it fails
 */
JNIEXPORT jlong JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeOpen (JNIEnv *env, jobject jobj, jstring port_name)
   {
   int comFd = open_com_path (env, port_name );
   
   if ( comFd < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "nativeOpen", "open FAIL", errno );
      return 0;
      }

   if ( ! isatty(comFd) )
      {
      throw_exception( env, IO_EXCEPTION, "nativeOpen", "isatty() == FALSE", ERROR_NULL );
      close(comFd);
      return 0;
      }

   struct ComStatus *comP = malloc (sizeof(struct ComStatus));

   if ( comP == NULL )
      {
      throw_exception( env, IO_EXCEPTION, "nativeOpen", "Cannot malloc stuct ComStatus", ERROR_NULL );
      close(comFd);
      return 0;
      }


   memset (comP,0,sizeof(*comP));

   comP->myselfPtr = comP;
   comP->portFd    = comFd;

   setNonblock (comP, FALSE);

   return (jlong)comP;
   }



/**
 * Close a port
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeClose (JNIEnv *env, jobject jobj, jlong comPtr)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);
   
   if ( comP == NULL ) return;

   if ( mustLog(DEBUG_CLOSE_PORT)) printf("nativeClose: CALL\n");

   if ( setNonblock (comP, TRUE) < 0 )
      {
	  // make sure that I will not get stuck on closing
	  throw_exception( env, IO_EXCEPTION, "nativeClose", "setNonblock FAIL", errno );
      }

   if ( tcflush ( comP->portFd, TCIOFLUSH ) < 0 )
      {
      // flush to avoid close hanging
      printf("nativeClose: tcflush error %d \n",errno);
      }

   if ( close(comP->portFd ) < 0 )
      {
	  // returns -1 on error
	  throw_exception( env, IO_EXCEPTION, "nativeClose", "close FAIL", errno );
      }
   
   // in any case free the referenced memory
   free (comP);
   }


static void adjustParamRtscts ( JNIEnv *env, struct termios *termP, jint jvalue )
   {
   switch ( jvalue )
      {
	  case com_engidea_comm_SerialPort_FLOWCONTROL_NONE:
		termP->c_cflag &= ~CRTSCTS;
		break;

	  case com_engidea_comm_SerialPort_FLOWCONTROL_RTSCTS_IN:
		 termP->c_cflag |= CRTSCTS;
		 break;

	  case com_engidea_comm_SerialPort_FLOWCONTROL_RTSCTS_OUT:
		termP->c_cflag |= CRTSCTS;
		break;

	 default:
		throw_exception( env, IO_EXCEPTION, "adjustParamRtscts", "UNSUPPORTED value for FLOWCONTROL", 0);
		break;
      }
   }

JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeSetPortParam (JNIEnv *env, jobject jobj, jlong comPtr, jint jwhat, jint jvalue)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return;

   struct termios *termP = &comP->termios;
   
   if ( tcgetattr(comP->portFd, termP) < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "SetPortParam", "tcgetattr FAIL", errno );
      return;
      }

   switch ( jwhat )
      {
      case com_engidea_linjcom_LinjcomPort_DCB_FLOWCONTROL:
    	  adjustParamRtscts (env, termP, jvalue );
    	  break;

      default:
         throw_exception( env, ILLEGAL_EXCEPTION, "SetPortParam", "unsupported jwhat", ERROR_NULL);
         break;
      }

   if ( tcsetattr(comP->portFd,TCSANOW,termP) < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "SetPortParam", "tcsetattr FAIL", errno );
      return;
      }
   }


static speed_t baud_to_com_speed (JNIEnv *env , int baudrate )
   {
   switch ( baudrate )
      {
      case 300:    return B300;
      case 600:    return B600;
      case 1200:   return B1200;
      case 2400:   return B2400;
      case 4800:   return B4800;
      case 9600:   return B9600;
      case 19200:  return B19200;
      case 38400:  return B38400;
      case 115200: return B115200;

      default:
    	  throw_exception( env, ILLEGAL_EXCEPTION, "baud_to_com_speed", "unsupported speed: set to 9600", 0);
    	  return B9600;
	  }
   }

static int com_speed_to_baud (JNIEnv *env , speed_t com_speed )
   {
   switch ( com_speed )
      {
      case B300:    return 300;
      case B600:    return 600;
      case B1200:   return 1200;
      case B2400:   return 2400;
      case B4800:   return 4800;
      case B9600:   return 9600;
      case B19200:  return 19200;
      case B38400:  return 38400;
      case B115200: return 115200;

      default:
    	  throw_exception( env, ILLEGAL_EXCEPTION, "com_speed_to_baud", "unsupported speed: return 9600", 0);
    	  return 9600;
	  }
   }



JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeGetPortParam (JNIEnv *env , jobject jobj, jlong comPtr, jint jwhat)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   struct termios *termP = &comP->termios;
 
   if ( tcgetattr(comP->portFd, termP) < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "GetPortParam", "tcgetattr FAIL", errno );
      return 0;
      }

   switch ( jwhat )
   	  {
   	  case com_engidea_linjcom_LinjcomPort_DCB_BAUDRATE:
         // always return the input speed
         return com_speed_to_baud (env, cfgetispeed(termP));

   	  case com_engidea_linjcom_LinjcomPort_DCB_DATABITS:
         // TODO
         return 8;

   	  case com_engidea_linjcom_LinjcomPort_DCB_PARITY:
	     // TODO
         return 0;

   	  case com_engidea_linjcom_LinjcomPort_DCB_STOPBITS:
         return 1;

   	  default:
         throw_exception( env, IO_EXCEPTION, "ILLEGAL_EXCEPTION", "Unmanaged jwhat", 0 );
         return 0;
   	  }
   }


JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeSetCommTimeout (JNIEnv *env, jobject job, jlong comPtr, jint timeoutMilli )
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return;

   int timeoutDs = timeoutMilli / 100;

   if ( timeoutDs < 0 )
	   timeoutDs = 0;
   else if ( timeoutDs > DS_COM_TIMEOUT_MAX )
	   timeoutDs = DS_COM_TIMEOUT_MAX;

   comP->timeoutDs = timeoutDs;

   struct termios *termP = &comP->termios;

   if ( timeoutDs > 0 )
   	   {
	   // this is a normal read with timeout
	   termP->c_cc[VTIME] = timeoutDs;
	   termP->c_cc[VMIN]  = 0;

	   // now prepare the setting for a select timeout
	   int seconds = timeoutDs / 10;

	   comP->writeTimeval.tv_sec = seconds;

	   int restoDs = timeoutDs % 10;

	   comP->writeTimeval.tv_usec = restoDs * 100000;
   	   }
   else
	   {
	   // this will block until interrupt or a char is read
	   termP->c_cc[VTIME] = 0;
	   termP->c_cc[VMIN]  = 1;
	   memset(&comP->writeTimeval,0,sizeof(comP->writeTimeval));
	   }

   if ( tcsetattr(comP->portFd,TCSANOW,termP) < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "nativeSetCommTimeout", "tcsetattr FAIL", errno );
      return;
      }
   }

JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeGetReceiveTimeout (JNIEnv *env, jobject jobj, jlong comPtr)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   int timeoutMilli = comP->termios.c_cc[VTIME] * 100;

   return timeoutMilli;
   }


/**
 * TIOCM_RNG, TIOCM_DSR, TIOCM_CD, and TIOCM_CTS
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeNotifyEnable (JNIEnv *env , jobject jobj, jlong comPtr, jint what, jboolean status)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return;

   switch ( what )
      {
   	  case com_engidea_linjcom_LinjcomPort_EV_CTS:
         if ( status )
        	 comP->notifyEnableBits |= TIOCM_CTS;
         else
        	 comP->notifyEnableBits &= ~TIOCM_CTS;
         break;

   	  case com_engidea_linjcom_LinjcomPort_EV_DSR:
         if ( status )
        	 comP->notifyEnableBits |= TIOCM_DSR;
         else
        	 comP->notifyEnableBits &= ~TIOCM_DSR;
         break;

   	  case com_engidea_linjcom_LinjcomPort_EV_RING:
         if ( status )
        	 comP->notifyEnableBits |= TIOCM_RNG;
         else
        	 comP->notifyEnableBits &= ~TIOCM_RNG;
         break;

   	  case com_engidea_linjcom_LinjcomPort_EV_RLSD:
         if ( status )
        	 comP->notifyEnableBits |= TIOCM_CD;
         else
        	 comP->notifyEnableBits &= ~TIOCM_CD;
         break;
      }
   }


JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeGetCommEvent (JNIEnv *env, jobject jobj, jlong comPtr)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   if ( ioctl(comP->portFd,TIOCMIWAIT,comP->notifyEnableBits) < 0 )
      {
      throw_exception( env, IO_EXCEPTION, "nativeGetCommEvent", "ioctl TIOCMIWAIT FAIL", errno );
      return 0;
      }

   // TODO should find out what is set using TIOCGICOUNT

   return 0;
   }



/**
 * return true if all is fine, FALSE otherwise
 */
static int convParityToTerm (JNIEnv *env, struct termios *termP, int parity )
   {
   if ( mustLog(DEBUG_POPARAMS) ) printf("convParityToTerm: parity=%d\n",parity);

   switch ( parity )
      {
      case com_engidea_comm_SerialPort_PARITY_NONE:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convParityToTerm: NOPARITY\n");
         termP->c_cflag &= ~PARENB;
         break;
      
      case com_engidea_comm_SerialPort_PARITY_MARK:
         throw_exception(env,ILLEGAL_EXCEPTION,"convParityToTerm","MARK Parity not supported",0);
         break;

      case com_engidea_comm_SerialPort_PARITY_EVEN:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convParityToTerm: EVENPARITY\n");
         termP->c_cflag |= PARENB;
         termP->c_cflag &= ~PARODD;
         break;

      case com_engidea_comm_SerialPort_PARITY_ODD:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convParityToTerm: ODDPARITY\n");
         termP->c_cflag |= PARENB;
         termP->c_cflag |= PARODD;
         break;

      case com_engidea_comm_SerialPort_PARITY_SPACE:
         throw_exception(env,ILLEGAL_EXCEPTION,"convParityToTerm","SPACE Parity not supported",0);
         break;

      default:
         throw_exception(env,ILLEGAL_EXCEPTION,"convParityToTerm","Illegal Parity value",0);
         return FALSE;
      }

   return TRUE;
   }

static int convStopbitToTerm (JNIEnv *env, struct termios *termP, int stopbit )
   {
   if ( mustLog(DEBUG_POPARAMS) ) printf("convStopbitToWin: stopbit=%d\n",stopbit);

   switch ( stopbit )
      {
      case com_engidea_comm_SerialPort_STOPBITS_1:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convStopbitToWin: ONESTOPBIT\n");
         termP->c_cflag &= ~CSTOPB;
         break;
      
      case com_engidea_comm_SerialPort_STOPBITS_2:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convStopbitToWin: TWOSTOPBITS\n");
         termP->c_cflag |= CSTOPB;
         break;

      default:
         throw_exception(env,ILLEGAL_EXCEPTION,"SetSerialPortParams","Illegal Stopbit",0);
         return FALSE;
      }

   return TRUE;
   }


static int convDatabitsToTerm (JNIEnv *env, struct termios *termP, int databits )
   {
   if ( mustLog(DEBUG_POPARAMS) ) printf("convDatabitsToTerm: databits=%d\n",databits);

   // clear the size mask
   termP->c_cflag &= ~CSIZE;

   switch ( databits )
      {
      case com_engidea_comm_SerialPort_DATABITS_5:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convDatabitsToTerm: 5 data\n");
         termP->c_cflag |= CS5;
         break;

      case com_engidea_comm_SerialPort_DATABITS_6:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convDatabitsToTerm: 6 data\n");
         termP->c_cflag |= CS6;
         break;

      case com_engidea_comm_SerialPort_DATABITS_7:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convDatabitsToTerm: 7 data\n");
         termP->c_cflag |= CS7;
         break;

   	  case com_engidea_comm_SerialPort_DATABITS_8:
         if ( mustLog(DEBUG_POPARAMS) ) printf("convDatabitsToTerm: 8 data\n");
         termP->c_cflag |= CS8;
         break;

      default:
         throw_exception(env,ILLEGAL_EXCEPTION,"convDatabitsToTerm","Illegal databits",0);
         return FALSE;
      }

   return TRUE;
   }


/**
 * After open you have to call this one to set port params
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeSetSerialPortParams (JNIEnv *env, jobject jobj, jlong comPtr, jint baudrate, jint databits, jint stopbits, jint parity)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return;
 
   if ( mustLog(DEBUG_POPARAMS) ) printf("nativeSetSerialPortParams: baud(%d)\n",baudrate);

   struct termios *termP = &comP->termios;

   if ( tcgetattr(comP->portFd, termP) < 0 )
      {
      throw_exception(env,IO_EXCEPTION,"SetSerialPortParams","tcgetattr FAIL", errno );
      return;
      }

   // strip off input processing
   termP->c_iflag &= ~(IGNBRK | BRKINT | ICRNL | INLCR | PARMRK | INPCK | ISTRIP | IXON);

   // no output processing either
   termP->c_oflag = 0;

   // no line processing
   termP->c_lflag &= ~(ECHO | ECHONL | ICANON | IEXTEN | ISIG);

   // assume I want RTS CTS and hang hup on close
   termP->c_cflag |= (CRTSCTS | HUPCL | CLOCAL | CREAD);

   // see http://www.unixwiz.net/techtips/termios-vmin-vtime.html
   // to have a pure timed read the VMIN must be zero
   termP->c_cc[VMIN]  = 0;
   termP->c_cc[VTIME] = 10;

   cfsetspeed (termP, baud_to_com_speed(env,baudrate));

   if ( ! convParityToTerm(env,termP,parity)  ) return;

   if ( ! convStopbitToTerm(env,termP,stopbits)  ) return;

   if ( ! convDatabitsToTerm(env,termP,databits)) return;

   if ( tcsetattr(comP->portFd, TCSANOW, termP) )
      {
      // Can happen, for way too many reasons
      throw_exception(env,IO_EXCEPTION,"SetSerialPortParams","tcsetattr FAIL", errno);
      }
   }


/**
 * Used to set/clear port lines
 * Eg: set DTR or something like that
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeSendEscape (JNIEnv *env, jobject jobj, jlong comPtr, jint what)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);
   int ioctlarg;

   if ( comP == NULL ) return;

   switch ( what )
      {
   	  case com_engidea_linjcom_LinjcomPort_ESC_SETDTR:
   		  ioctlarg = TIOCM_DTR;
   		  if ( ioctl (comP->portFd, TIOCMBIS, &ioctlarg ) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCMBIS TIOCM_DTR FAIL", errno);
         break;

   	  case com_engidea_linjcom_LinjcomPort_ESC_CLRDTR:
   		  ioctlarg = TIOCM_DTR;
   		  if ( ioctl (comP->portFd, TIOCMBIC, &ioctlarg ) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCMBIC TIOCM_DTR FAIL", errno);
         break;

   	  case com_engidea_linjcom_LinjcomPort_ESC_SETRTS:
  		  ioctlarg = TIOCM_RTS;
   		  if ( ioctl (comP->portFd, TIOCMBIS, &ioctlarg ) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCMBIS TIOCM_RTS FAIL", errno);
         break;

   	  case com_engidea_linjcom_LinjcomPort_ESC_CLRRTS:
  		  ioctlarg = TIOCM_RTS;
   		  if ( ioctl (comP->portFd, TIOCMBIC, &ioctlarg ) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCMBIC TIOCM_RTS FAIL", errno);
         break;

   	  case com_engidea_linjcom_LinjcomPort_ESC_SETBREAK:
  		  ioctlarg = TIOCSBRK;
   		  if ( ioctl (comP->portFd, TIOCMBIS, &ioctlarg) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCSBRK FAIL", errno);
   		  break;

   	  case com_engidea_linjcom_LinjcomPort_ESC_CLRBREAk:
  		  ioctlarg = TIOCSBRK;
   		  if ( ioctl (comP->portFd, TIOCMBIC, &ioctlarg) < 0 )
             throw_exception(env,IO_EXCEPTION,"nativeSendEscape","ioctl TIOCCBRK FAIL", errno);
   		  break;

   	  default:
         throw_exception(env,ILLEGAL_EXCEPTION,"nativeSendEscape","EscapeCommFunction UNSUPPORTED", 0);
         break;
      }
   }


/**
 * @return the number of bytes written
 * On timeout the number of bytes written will be LESS that what has been given !
 */
JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeWriteArray (JNIEnv * env, jobject jobj, jlong comPtr, jbyteArray jarray, jint joffset, jint jlen)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   if ( mustLog(DEBUG_WRITE) ) printf("writeArray: length=%d\n",jlen);

   if ( comP->timeoutDs > 0 )
   	  {
	  struct timeval tv;
	  fd_set wfds;

	  FD_ZERO(&wfds);
	  FD_SET(comP->portFd, &wfds);

	  tv = comP->writeTimeval;

	  int selres = select(comP->portFd+1, NULL, &wfds, NULL, &tv);

      // this is basically a timeout, no chars written
	  if ( selres == 0 ) return 0;

	  if ( selres < 0 )
		  {
		  throw_exception(env,IO_EXCEPTION,"nativeWriteArray","select failed", errno);
		  return -1;
		  }

	   if ( ! FD_ISSET(comP->portFd, &wfds))
		  {
		  // this is an extra check on select behavior
		  throw_exception(env,IO_EXCEPTION,"nativeWriteArray","impossible ! FD_ISSET", 0);
		  return -1;
		  }
   	  }

   // get the data to write
   jbyte *dataP = (*env)->GetByteArrayElements( env, jarray, 0 );

   // this will return the number of chars written
   int retcode = write(comP->portFd, dataP+joffset, jlen);

   // release the buffer without copying back anything
   (*env)->ReleaseByteArrayElements( env, jarray, dataP, JNI_ABORT );

   return retcode;
   }


/**
 * Read data from port in overlapped mode.
 * Note that on timeout you may actually have read some data !
 */
JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeReadArray (JNIEnv *env, jobject jobj, jlong comPtr, jbyteArray jarray, jint joffset, jint jlen)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   jbyte *dataP = (*env)->GetByteArrayElements( env, jarray, 0 );

   if ( mustLog(DEBUG_READ) ) printf("readArray: len=%d\n",jlen);

   int letti = read(comP->portFd, dataP + joffset, jlen);

   // tell java to pick up the new data, it MAY be less than what I wanted if I have Timeout or error
   (*env)->ReleaseByteArrayElements( env, jarray, dataP, 0 );

   // Generally speaking I cannot have an EOF (I think) on a serial port.
   // So, basically, when letti != jlen it is a timeout....

   if ( letti < 0 )
      {
      // it is possible that read fails, beside timeout. let's report it
	  throw_exception(env,IO_EXCEPTION,"nativeReadArray"," read error", errno);
	  }

   return letti; 
   }


JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeGetModemStatus ( JNIEnv *env,   jobject jobj, jlong comPtr )
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   int32 status;

   if ( ioctl(comP->portFd, TIOCMGET, &status) < 0 )
      {
      throw_exception(env,IO_EXCEPTION,"nativeGetModemStatus","ioctl TIOCMGET FAIL", errno);
      return 0;
      }

   int32 risul = 0;

   if (status & TIOCM_CTS ) risul |= com_engidea_linjcom_LinjcomPort_MDM_CTS_ON;

   if (status & TIOCM_RNG ) risul |= com_engidea_linjcom_LinjcomPort_MDM_RING_ON;

   if (status & TIOCM_DSR  ) risul |= com_engidea_linjcom_LinjcomPort_MDM_DSR_ON;

   if (status & TIOCM_CAR  ) risul |= com_engidea_linjcom_LinjcomPort_MDM_CD_ON;

   if (status & TIOCM_DTR  ) risul |= com_engidea_linjcom_LinjcomPort_MDM_DTR_ON;

   if (status & TIOCM_RTS  ) risul |= com_engidea_linjcom_LinjcomPort_MDM_RTS_ON;

   return risul;
   }


/**
 * Flush possible pending chars BOTH RX AND TX !!!
 */
JNIEXPORT void JNICALL Java_com_engidea_linjcom_LinjcomPort_nativePurgeComm (JNIEnv *env , jobject jobj, jlong comPtr)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return;

   if ( tcflush ( comP->portFd, TCIOFLUSH ) < 0 )
      {
	  throw_exception(env,IO_EXCEPTION,"nativePurgeComm","tcflush FAIL", errno);
      }
   }


JNIEXPORT jint JNICALL Java_com_engidea_linjcom_LinjcomPort_nativeAvailable (JNIEnv *env, jobject jobj, jlong comPtr)
   {
   struct ComStatus *comP = validateComPtr (env, comPtr);

   if ( comP == NULL ) return 0;

   int bytes_avaiable = 0;

   if ( ioctl(comP->portFd, FIONREAD, &bytes_avaiable) < 0 )
      {
	  throw_exception(env,IO_EXCEPTION,"nativePurgeComm","ioctl FIONREAD FAIL", errno);
	  }

   return bytes_avaiable;
   }




JNIEXPORT jint JNICALL JNI_OnLoad(JavaVM *java_vm, void *reserved)
   {
//   printf("JNI_OnLoad\n");
   return JNI_VERSION_1_2;  
   }


JNIEXPORT void JNICALL JNI_OnUnload(JavaVM *vm, void *reserved)
   {
//   printf("JNI_OnUnload\n");
   }

